﻿using System;
using System.Collections.Generic;
using System.Linq;
using Clarion;
using Clarion.Framework;
using Clarion.Framework.Core;
using Clarion.Framework.Templates;
using WorldServerLibrary.Model;
using System.Threading;

namespace ClarionSimulation
{
    /// <summary>

    /// This event will be fired at the beginning of each cognitive cycle in order to get the current visual sensor information.
    /// </summary>
    /// <returns>The Visual Sensor Information Itens</returns> 
    public delegate IList<Thing> InputVisualSensorialInformationEventHandler();

    /// <summary>
    /// This event will be fired at the end of each cognitive cycle in order to translate the selected action to a World Server Command
    /// </summary>
    /// <param name="externalAction">The external action that was choice</param>
    public delegate void OutputActionChunkEventHandler(ClarionAgentActionType externalAction);

    /// <summary>
    /// Public enum that represents all possibilities of agent actions
    /// </summary>
    public enum ClarionAgentActionType
    {
        DO_NOTHING,
        ROTATE_CLOCKWISE,
        GO_AHEAD,
        LOOK_FOR_SOMETHING, // ra036647 - if no food and no jewels are seen, avoid to move to not consume too much energy, prefer to rotate
        GO_TO_FOOD, // ra036647 - if a food is subjected
        GO_TO_JEWEL, // ra036647 - if a jewel is subjected
        GET_SOMETHING, // ra036647 - if food then eat else grab it (jewel) because "you are close enough to do it"
    }

    public class ClarionAgent
    {
        #region Constants

        const int FUEL_WARNING_LEVEL = 951; // ra036647 - fuel warning level

        /// <summary>
        /// Constant that represents the Visual Sensor
        /// </summary>
        private String SENSOR_VISUAL_DIMENSION = "VisualSensor";

        /// <summary>
        /// Constant that represents that there is at least one wall ahead
        /// </summary>
        private String DIMENSION_WALL_AHEAD = "WallAhead";

        #region ra036647
        private String DIMENSION_JEWEL_AHEAD = "JewelAhead";
        private String DIMENSION_FOOD_AHEAD = "FoodAhead";
        private String DIMENSION_NOTHING_AHEAD = "NothingAhead";
        private String DIMENSION_SOMETHING_READY = "SomethingReady";
        #endregion

        #endregion

        #region Properties

        #region Simulation

        /// <summary>
        /// If this value is greater than zero, the agent will have a finite number of cognitive cycle. Otherwise, it will have infinite cycles.
        /// </summary>
        public double MaxNumberOfCognitiveCycles = -1;

        /// <summary>
        /// Current cognitive cycle number
        /// </summary>
        private double CurrentCognitiveCycle = 0;

        /// <summary>
        /// Time between cognitive cycle in miliseconds
        /// </summary>
        public Int32 TimeBetweenCognitiveCycles = 0;

        /// <summary>
        /// A thread Class that will handle the simulation process
        /// </summary>
        private Thread runThread;

        #endregion

        #region Agent

        /// <summary>
        /// The agent 
        /// </summary>
        private Clarion.Framework.Agent CurrentAgent;

        /// <summary>
        /// For each cognitive cycle, this event will be called in order to the agent receives the current sensorial information
        /// </summary>
        public event InputVisualSensorialInformationEventHandler OnNewVisualSensorialInformation;

        /// <summary>
        /// For each cognitive cycle, this event will be called when the agent selects one action
        /// </summary>
        public event OutputActionChunkEventHandler OnNewExternalActionSelected;

        double _max = double.MaxValue; // ra036647 - shortcut
        double _min = double.MinValue; // ra036647 - shortcut

        #endregion

        #region Perception Input

        /// <summary>
        /// Perception input to indicates a wall ahead
        /// </summary>
        private DimensionValuePair InputWallAhead;

        #region ra036647
        private DimensionValuePair InputJewelAhead;
        private DimensionValuePair InputFoodAhead;
        private DimensionValuePair InputNothingAhead;
        private DimensionValuePair InputSomethingReady;
        #endregion

        #endregion

        #region Action Output

        /// <summary>
        /// Output action that makes the agent to rotate clockwise
        /// </summary>
        private ExternalActionChunk OutputRotateClockwise;

        /// <summary>
        /// Output action that makes the agent go ahead
        /// </summary>
        private ExternalActionChunk OutputGoAhead;

        #region ra036647
        private ExternalActionChunk OutputLookForSomething;
        private ExternalActionChunk OutputGoToJewel;
        private ExternalActionChunk OutputGoToFood;
        private ExternalActionChunk OutputGetSomething;
        #endregion

        #endregion

        #endregion

        #region Constructor

        public ClarionAgent()
        {
            // Initialize the agent
            CurrentAgent = World.NewAgent("Current Agent");

            this._max = CurrentAgent.Parameters.MAX_ACTIVATION; // ra036647 - shortcut
            this._min = CurrentAgent.Parameters.MIN_ACTIVATION; // ra036647 - shortcut

            // Initialize Input Information
            InputWallAhead = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_WALL_AHEAD);

            // Initialize Output actions
            OutputRotateClockwise = World.NewExternalActionChunk(ClarionAgentActionType.ROTATE_CLOCKWISE.ToString());
            OutputGoAhead = World.NewExternalActionChunk(ClarionAgentActionType.GO_AHEAD.ToString());

            #region ra036647
            this.InputJewelAhead = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_JEWEL_AHEAD);
            this.InputFoodAhead = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_FOOD_AHEAD);
            this.InputNothingAhead = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_NOTHING_AHEAD);
            this.InputSomethingReady = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_SOMETHING_READY);
            this.OutputLookForSomething = World.NewExternalActionChunk(ClarionAgentActionType.LOOK_FOR_SOMETHING.ToString());
            this.OutputGoToJewel = World.NewExternalActionChunk(ClarionAgentActionType.GO_TO_JEWEL.ToString());
            this.OutputGoToFood = World.NewExternalActionChunk(ClarionAgentActionType.GO_TO_FOOD.ToString());
            this.OutputGetSomething = World.NewExternalActionChunk(ClarionAgentActionType.GET_SOMETHING.ToString());
            #endregion

            //Create thread to simulation
            runThread = new Thread(RunThread);
			Console.WriteLine("Agent started");
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Run the Simulation in World Server 3d Environment
        /// </summary>
        public void Run()
        {                
			Console.WriteLine ("Running ...");
            // Setup Agent to run
            if (runThread != null && !runThread.IsAlive)
            {
                SetupAgentInfraStructure();

                if (OnNewVisualSensorialInformation != null)
                {
                    // Start Simulation Thread                
                    runThread.Start(null);
                }
            }
        }

        /// <summary>
        /// Abort the current Simulation
        /// </summary>
        /// <param name="deleteAgent">If true beyond abort the current simulation it will die the agent.</param>
        public void Abort(Boolean deleteAgent)
        {   Console.WriteLine ("Aborting ...");
            if (runThread != null && runThread.IsAlive)
            {
                runThread.Abort();
            }

            if (CurrentAgent != null && deleteAgent)
            {
                CurrentAgent.Die();
            }
        }

        #endregion

        #region Setup Agent Methods

        /// <summary>
        /// Setup agent infra structure (ACS, NACS, MS and MCS)
        /// </summary>
        private void SetupAgentInfraStructure()
        {
            // Setup the ACS Subsystem
            SetupACS();                    
        }

        private void SetupMS()
        {            
            //RichDrive
        }

        /// <summary>
        /// Setup the ACS subsystem
        /// </summary>
        private void SetupACS()
        {
            // Create Colission Wall Rule
            SupportCalculator avoidCollisionWallSupportCalculator = FixedRuleDelegateToAvoidColissionWall;
            FixedRule ruleAvoidCollisionWall = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputRotateClockwise, avoidCollisionWallSupportCalculator);

            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit(ruleAvoidCollisionWall);

            // Create Colission To Go Ahead
            SupportCalculator goAheadSupportCalculator = FixedRuleDelegateToGoAhead;
            FixedRule ruleGoAhead = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputGoAhead, goAheadSupportCalculator);
            
            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit(ruleGoAhead);

            #region ra036647
            this.CurrentAgent.Commit(
                AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputLookForSomething, (SupportCalculator) FixedRuleLookForSomethingSupportCalculator)
            );
            this.CurrentAgent.Commit(
                AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputGoToFood, (SupportCalculator)FixedRuleGoToFoodSupportCalculator)
            );
            this.CurrentAgent.Commit(
                AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputGoToJewel, (SupportCalculator)FixedRuleGoToJewelSupportCalculator)
            );
            this.CurrentAgent.Commit(
                AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputGetSomething, (SupportCalculator)FixedRuleGetSomethingSupportCalculator)
            );
            #endregion

            // Disable Rule Refinement
            CurrentAgent.ACS.Parameters.PERFORM_RER_REFINEMENT = false;

            // The selectio type will be probabilistic
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_METHOD = ActionCenteredSubsystem.LevelSelectionMethods.STOCHASTIC;

            // The selection of the action will be fixed (not variable) i.e. only the statement defined above.
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_OPTION = ActionCenteredSubsystem.LevelSelectionOptions.FIXED;

            // Define Probabilistic values
            CurrentAgent.ACS.Parameters.FIXED_FR_LEVEL_SELECTION_MEASURE = 1;
            CurrentAgent.ACS.Parameters.FIXED_IRL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_BL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_RER_LEVEL_SELECTION_MEASURE = 0;
        }

        /// <summary>
        /// Make the agent perception. In other words, translate the information that came from sensors to a new type that the agent can understand
        /// </summary>
        /// <param name="sensorialInformation">The information that came from server</param>
        /// <returns>The perceived information</returns>
        private SensoryInformation MakePerceptionFromSensorialInput(IList<Thing> sensorialInformation)
        {
            // New sensory information
            SensoryInformation si = World.NewSensoryInformation(CurrentAgent);
            #region não importante

            /*
            // Detect if we have a wall ahead
            Boolean wallAhead = sensorialInformation.Where(item => (item.CategoryId == Thing.CATEGORY_BRICK && item.DistanceToCreature <= 61)).Any();
            double wallAheadActivationValue = wallAhead ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION;
            si.Add(InputWallAhead, wallAheadActivationValue);
            //Console.WriteLine(sensorialInformation);
            */
            #endregion
            Creature c = (Creature)sensorialInformation.Where(item => (item.CategoryId == Thing.CATEGORY_CREATURE)).First();
            //c.PrintLeaflet();
            
            
            var myCreature = (Creature)sensorialInformation.First(i => i.CategoryId == Thing.CATEGORY_CREATURE && i.DistanceToCreature <= 0.0 && i.comX <= 0.0 && i.comY <= 0.0);
            var fuel = myCreature.Fuel;

            double nothing = _min, food = _min, jewel = _min, something = _min;

            bool somethingToGet = sensorialInformation.Where(i => Thing.foodOrJewel.Contains(i.CategoryId) && i.DistanceToCreature <= 22.0).Any();
            bool somethingIsSeen = sensorialInformation.Where(i => Thing.foodOrJewel.Contains(i.CategoryId)).Any();
            
            var foodsList = sensorialInformation.Where(i => Thing.foods.Contains(i.CategoryId));
            bool foodIsSeen = foodsList.Any();

            var jewelsList = sensorialInformation.Where(i => Thing.jewels.Contains(i.CategoryId));
            bool jewelIsSeen = jewelsList.Any();

            if (somethingToGet) something = _max;
            else if (!somethingIsSeen) nothing = _max;
            else
            {
                if (jewelIsSeen && foodIsSeen)
                    if (fuel < FUEL_WARNING_LEVEL) food = _max;
                    else jewel = _max;
                else
                {
                    food = (foodIsSeen ? _max : _min);
                    jewel = (jewelIsSeen ? _max : _min);
                }
            }

            si.Add(InputSomethingReady, something);
            si.Add(InputNothingAhead, nothing);
            si.Add(InputFoodAhead, food);
            si.Add(InputJewelAhead, jewel);

            return si;
        }

        #endregion

        #region Delegate Methods

        #region Fixed Rules

        private double FixedRuleDelegateToAvoidColissionWall(ActivationCollection currentInput, Rule target)
        {
            // See partial match threshold to verify what are the rules available for action selection
            return ((currentInput.Contains(InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
        }

        private double FixedRuleDelegateToGoAhead(ActivationCollection currentInput, Rule target)
        {
            // Here we will make the logic to go ahead
            return ((currentInput.Contains(InputWallAhead, CurrentAgent.Parameters.MIN_ACTIVATION))) ? 1.0 : 0.0;
        }

        #region ra036647
        private double FixedRuleLookForSomethingSupportCalculator(ActivationCollection currentInput, Rule target)
        {
            return ((currentInput.Contains(InputNothingAhead, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
        }
        private double FixedRuleGoToFoodSupportCalculator(ActivationCollection currentInput, Rule target)
        {
            return ((currentInput.Contains(InputFoodAhead, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
        }
        private double FixedRuleGoToJewelSupportCalculator(ActivationCollection currentInput, Rule target)
        {
            return ((currentInput.Contains(InputJewelAhead, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
        }
        private double FixedRuleGetSomethingSupportCalculator(ActivationCollection currentInput, Rule target)
        {
            return ((currentInput.Contains(InputSomethingReady, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
        }
        #endregion

        #endregion

        #region Run Thread Method

        private void RunThread(object obj)
        {
			Console.WriteLine("Starting Cognitive Cycle ... press CTRL-C to finish !");
            // Cognitive Cycle starts here getting sensorial information
            while (CurrentCognitiveCycle != MaxNumberOfCognitiveCycles)
            {   
                // Get current sensorial information                    
                IList<Thing> sensorialInformation = OnNewVisualSensorialInformation();

                // Make the perception
                SensoryInformation si = MakePerceptionFromSensorialInput(sensorialInformation);

                //Perceive the sensory information
                CurrentAgent.Perceive(si);

                //Choose an action
                ExternalActionChunk chosen = CurrentAgent.GetChosenExternalAction(si);

                // Get the selected action
                String actionLabel = chosen.LabelAsIComparable.ToString();
                ClarionAgentActionType actionType = (ClarionAgentActionType)Enum.Parse(typeof(ClarionAgentActionType), actionLabel, true);

                // Call the output event handler
                if (OnNewExternalActionSelected != null)
                {
                    OnNewExternalActionSelected(actionType);
                }

                // Increment the number of cognitive cycles
                CurrentCognitiveCycle++;

                //Wait to the agent accomplish his job
                if (TimeBetweenCognitiveCycles > 0)
                {
                    Thread.Sleep(TimeBetweenCognitiveCycles);
                }
            }
        }

        #endregion

        #endregion
    }
}
