
using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Threading;
using ClarionSimulation;
using WorldServerLibrary;
using WorldServerLibrary.Model;
using WorldServerLibrary.Exceptions;

namespace ClarionDEMO
{
	class MainClass
	{
		#region properties
		private WorldServer worldServer = null;
        private ClarionAgent agent;
        String creatureId = String.Empty;
        String creatureName = String.Empty;
		double prad = 0;
        const long debugInMilliseconds = 1000; // ra036647 - DEBUG control
        DateTime lastDebug = DateTime.Now; // ra036647 - DEBUG control
        Thing closestFood = null; // ra036647 - mapping
        Thing targetJewel = null; // ra036647 - mapping
        Thing somethingToGet = null; // ra036647 - mapping
		#endregion

		#region constructor
		public MainClass() {
            Console.WriteLine("Clarion Demo V0.4");
            //Console.WriteLine("Press a key to continue...");
            //Console.ReadKey();
			agent = new ClarionAgent();
            agent.OnNewVisualSensorialInformation += new InputVisualSensorialInformationEventHandler(agent_OnNewVisualSensorialInformation);
            agent.OnNewExternalActionSelected += new OutputActionChunkEventHandler(agent_OnNewExternalActionSelected);
			try
            {
                worldServer = new WorldServer("localhost", 4011);
                String message = worldServer.Connect();

                if (worldServer != null && worldServer.IsConnected)
                {
                    Console.Out.WriteLine ("[SUCCESS] " + message + "\n");
					//worldServer.SendWorldReset();
                    worldServer.NewCreature(500, 500, 0, out creatureId, out creatureName);
					worldServer.SendCreateLeaflet();
                    //worldServer.NewBrick(4, 747, 2, 800, 567);
                    //worldServer.NewBrick(4, 50, -4, 747, 47);
                    //worldServer.NewBrick(4, 49, 562, 796, 599);
                    //worldServer.NewBrick(4, -2, 6, 50, 599);

                    #region ra036647
                    //worldServer.NewFood(0, 300, 5);
                    #endregion

                    if (!String.IsNullOrWhiteSpace(creatureId))
                    {
                        worldServer.SendStartCamera(creatureId);
                        worldServer.SendStartCreature(creatureId);
                    }

                    Console.Out.WriteLine("Creature created with name: " + creatureId + "\n");
                    agent.Run();
					Console.Out.WriteLine("Running Simulation ...\n");
                }
            }
            catch (WorldServerInvalidArgument invalidArtgument)
            {
                Console.Out.WriteLine(String.Format("[ERROR] Invalid Argument: {0}\n", invalidArtgument.Message));
            }
            catch (WorldServerConnectionError serverError)
            {
                Console.Out.WriteLine(String.Format("[ERROR] Is is not possible to connect to server: {0}\n", serverError.Message));
            }
            catch (Exception ex)
            {
                Console.Out.WriteLine(String.Format("[ERROR] Unknown Error: {0}\n", ex.Message));
            }
		}
		#endregion

		#region Methods
		public static void Main (string[] args)	{
			new MainClass();
		}

        IList<Thing> agent_OnNewVisualSensorialInformation()
        {
            // ra036647 - extract worldserver data in visual space to all analysis
            IList<Thing> response = worldServer.SendGetCreatureState(creatureName);
            var sack = worldServer.SendGetSack(creatureId);

            var myCreature = (Creature) response.First(i => i.CategoryId == Thing.CATEGORY_CREATURE && i.Name == creatureName);
            var fuel = myCreature.Fuel;
            
            var consolidatedLeaflet = myCreature.getConsolidatedLeaflet();
            sack.update(consolidatedLeaflet);

            closestFood = response.Where(i => Thing.foods.Contains(i.CategoryId)).OrderBy(i => i.DistanceToCreature).FirstOrDefault();
            targetJewel = (from thing in response.Where(i => Thing.jewels.Contains(i.CategoryId))               // selects only jewels
                          let leaflet = consolidatedLeaflet.First(l => l.Key == thing.Material.Color).Value     // find leaflet consolidated for jewel's name
                          let remaining = leaflet.totalNumber - leaflet.collected                               // calculates remaining
                          let priority = (remaining <= 0) ? 0 : remaining                                       // calculates priority, zero if leaflet is completed
                          orderby priority descending, thing.DistanceToCreature ascending                       // rank jewels based on priority
                          select thing).FirstOrDefault();                                                       // gets the most required
            somethingToGet = response.Where(i => Thing.foodOrJewel.Contains(i.CategoryId) && i.DistanceToCreature <= 22.0).FirstOrDefault();

            // ra036647 - DEBUG
            if (lastDebug.AddMilliseconds(debugInMilliseconds) < DateTime.Now)
            {
                lastDebug = DateTime.Now;
                Console.WriteLine("FUEL = " + fuel);
                Print(consolidatedLeaflet);
            }
            
            if (worldServer != null && worldServer.IsConnected)
            {
				prad = (Math.PI / 180) * response.First().Pitch;
			    while (prad > Math.PI) prad -= 2 * Math.PI;
			    while (prad < - Math.PI) prad += 2 * Math.PI;

            }

            return response;
        }

        private void Print(Dictionary<string, LeafletItem> consolidatedLeaflet)
        {
            Dictionary<string, LeafletItem> l = consolidatedLeaflet;
            Console.Write("--> ");
			bool completed = true;
            foreach (var pair in l)
            {
				Console.Write(pair.Key + " (R=" + pair.Value.totalNumber + ";C=" + pair.Value.collected + ") ");
				if (pair.Value.totalNumber > pair.Value.collected) completed = false;
			}
			Console.WriteLine(" ");
			if (!completed) {
				Console.WriteLine ("LEAFTLET NOT COMPLETED YET");
			} else {
				Console.WriteLine ("LEAFTLET COMPLETED !!!");
			}
        }

        void agent_OnNewExternalActionSelected(ClarionAgentActionType externalAction)
        {
            Thread.CurrentThread.CurrentCulture = new CultureInfo("en-US");
            double newprad = prad + Math.PI / 1000;
            while (newprad > Math.PI) newprad -= 2 * Math.PI;
            while (newprad < -Math.PI) newprad += 2 * Math.PI;

            if (worldServer != null && worldServer.IsConnected)
            {
                switch (externalAction)
                {
                    case ClarionAgentActionType.DO_NOTHING:
                        // Do nothing as the own value says
                        break;
                    case ClarionAgentActionType.ROTATE_CLOCKWISE:
                    case ClarionAgentActionType.LOOK_FOR_SOMETHING: // ra036647
                        worldServer.SendSetAngle(creatureId, 0, 0, newprad);
                        break;
                    case ClarionAgentActionType.GO_AHEAD:
                        worldServer.SendSetAngle(creatureId, 1, 1, prad);
                        break;
                    case ClarionAgentActionType.GO_TO_FOOD:
                        if (closestFood != null)
                        {
                            worldServer.SendSetAngle(creatureId, 0, 0, prad);
                            worldServer.SendSetGoTo(creatureId, 1, 1, closestFood.X1, closestFood.Y1);
                        }
                        break;
                    case ClarionAgentActionType.GO_TO_JEWEL:
                        if (targetJewel != null)
                        {
                            worldServer.SendSetAngle(creatureId, 0, 0, prad);
                            worldServer.SendSetGoTo(creatureId, 1, 1, targetJewel.X1, targetJewel.Y1);
                        }
                        break;
                    case ClarionAgentActionType.GET_SOMETHING:
                        if (somethingToGet != null)
                        {
                            worldServer.SendSetAngle(creatureId, 0, 0, prad);
                            if (Thing.jewels.Contains(somethingToGet.CategoryId)) // if jewel, pick it
                            {
                                worldServer.SendSackIt(creatureId, somethingToGet.Name);
                            }
                            else if (Thing.foods.Contains(somethingToGet.CategoryId)) // if food, eat
                            {
                                worldServer.SendEatIt(creatureId, somethingToGet.Name);
                            }
                            else // otherwise, turn creature to run away because it's trapped
                            {
                                worldServer.SendSetAngle(creatureId, 0, 0, newprad);
                            }
                        }
                        break;
                    default:
                        break;
                }
            }
        }

        #endregion
	}
	
	
}
